"""Utilities."""

import os

def filtany(entities, **kw):
  """Filter a set of entities based on method return. Use keyword arguments.
  
  Example:
    filtmeth(entities, id='123')
    filtmeth(entities, name='bart')

  Multiple filters are 'OR'.
  """
  ret = set()
  for k,v in kw.items():
    for entity in entities:
      if getattr(entity, k)() == v:
        ret.add(entity)
  return ret

def filtfirst(entities, **kw):
  """Return the first matching entity, sorted by id()."""
  ret = sorted(filtany(entities, **kw), key=lambda x:x.id())
  if not ret:
    raise ValueError('No result')
  return ret[0]

##### Utilities for tests #####

def example_feed(feed='sample-feed.zip'):
  return os.path.join(
    os.path.dirname(__file__), 
    'examples',
    feed
    )
  
def preload_agency(**kw):
  import feed  
  agency_id = kw.pop('agency_id', 'DTA')
  f = feed.Feed(example_feed(**kw))
  f.preload()
  agency = f.agency(agency_id)
  return agency



"""GTFS Route entity."""
import collections

import entity
import geom
import validation

class Route(entity.Entity):
  """GTFS Route entity."""
  ENTITY_TYPE = 'r'
  KEY = 'route_id'
  REQUIRED = [
    'route_id',
    'route_short_name',
    'route_long_name',
    'route_type'    
  ]
  OPTIONAL = [
    'agency_id',
    'route_desc',
    'route_url',
    'route_color',
    'route_text_color'
  ]
  
  def name(self):
    return self.get('route_short_name') or self.get('route_long_name')

  def id(self):
    return self.get('route_id')

  def bbox(self):
    return geom.bbox(self.stops())

  def json(self):
    return {
      'type': 'Feature',
      'name': self.name(),
      'properties': self.data,
      'bbox': self.bbox(),
      'geometry': self.geometry()
    }

  def geometry(self):
    # Return a line for most popular shape_id or stop pattern 
    #   in each direction_id.
    d0 = collections.defaultdict(int)
    d1 = collections.defaultdict(int)
    # Grab the shapes.
    try:
      shapes = self._feed.shapes()
    except KeyError:
      shapes = {}      
    for trip in self.trips():
      if trip.get('shape_id') and trip.get('shape_id') in shapes:
        seq = tuple(shapes[trip['shape_id']].points())
      else:
        seq = tuple(i.point() for i in trip.stop_sequence())
      if int(trip.get('direction_id') or 0):
        d1[seq] += 1
      else:
        d0[seq] += 1
    # Sort to find the most popular shape.
    route0 = []
    if d0:
      route0 = sorted(d0.items(), key=lambda x:x[1])[-1][0]
    route1 = []
    if d1:
      route1 = sorted(d1.items(), key=lambda x:x[1])[-1][0]
    return {
      'type':'MultiLineString',
      'coordinates': [route0, route1]
    }
  
  def vehicle(self):
    return {
      '0': 'tram',
      '1': 'metro',
      '2': 'rail',
      '3': 'bus',
      '4': 'ferry',
      '5': 'cablecar',
      '6': 'gondola',
      '7': 'funicular',
      None: None,
      '': None
    }[self.get('route_type')]
  
  # Graph.  
  def trips(self):
    """Return trips for this route."""
    return set(self.children()) # copy

  def stops(self):
    """Return stops served by this route."""
    serves = set()
    for trip in self.trips():
      for stop_time in trip.stop_times():
        serves |= stop_time.stops()
    return serves

  ##### Validation #####
  def validate(self, validator=None):
    validator = super(Route, self).validate(validator)
    # Required
    with validator(self): 
      assert self.get('route_id'), "Required: route_id"
    with validator(self):
      assert self.get('route_type'), "Required: route_type"
      assert self.vehicle(), "Invalid route_type"
    with validator(self): 
      assert self.get('route_short_name') or self.get('route_long_name'), \
        "Must provide either route_short_name or route_long_name"
    # TODO: Warnings: 
    #   short name too long
    #   short name == long name
    #   route_desc != route name
    # Optional
    with validator(self):
      if self.get('agency_id'): pass
    with validator(self):
      if self.get('route_desc'): pass      
    with validator(self): 
      if self.get('route_url'):
        assert validation.valid_url(self.get('route_url')), \
          "Invalid route_url"
    with validator(self): 
      if self.get('route_color'):
        assert validation.valid_color(self.get('route_color')), \
          "Invalid route_color"
    with validator(self): 
      if self.get('route_text_color'):
        assert validation.valid_color(self.get('route_text_color')), \
          "Invalid route_text_color"
    return validator



"""GTFS Fare Attributes."""
import datetime

import entity
import geom
import util
import widetime
import validation

class FareAttribute(entity.Entity):
  KEY = 'fare_id'
  REQUIRED = [
    'fare_id',
    'price',
    'currency_type',
    'payment_method',
    'transfers',
  ]
  OPTIONAL = [
    'transfer_duration'
  ]
  
  def validate(self, validator=None):
    validator = super(FareAttribute, self).validate(validator)
    # Required
    with validator(self):
      assert self.get('fare_id'), "Required: fare_id"
    with validator(self):
      assert self.get('price'), "Required: price"
    with validator(self):
      assert validation.valid_float(self.get('price')), "Invalid price"
    with validator(self):
      assert self.get('currency_type'), "Required: currency_type"
    with validator(self):
      assert self.get('payment_method'), "Required: payment_method"
    with validator(self):
      assert validation.valid_bool(self.get('payment_method')), \
        "Invalid payment_method"
    with validator(self):
      assert validation.valid_int(self.get('transfers'), vmin=0, vmax=2, empty=True), \
        "Invalid transfers"
    # Optional
    with validator(self):
      if self.get('transfer_duration'):
        assert validation.valid_int(self.get('transfer_duration'), vmin=0), \
          "Invalid transfer_duration"
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(FareAttribute, self).validate_feed(validator)
    return validator


import unittest

import feed
import entities
import util
    
class TestTrip(unittest.TestCase):
  expect = {
    'route_id': 'CITY', 
    'trip_headsign': '', 
    'block_id': '', 
    'direction_id': '1', 
    'shape_id': '',
    'trip_id': 'CITY2', 
    'service_id': 'FULLW'
  }
  
  def test_stop_times(self):
    agency = util.preload_agency()
    entity = agency.trip(self.expect['trip_id'])
    assert len(entity.stop_times()) == 5

  def test_stop_sequence(self):
    agency = util.preload_agency()
    entity = agency.trip(self.expect['trip_id'])
    stop_sequence = entity.stop_sequence()
    assert len(stop_sequence) == 5
    expect = ['1', '2', '3', '4', '5']
    for i,j in zip(stop_sequence, expect):
      assert i.get('stop_sequence') == j



import unittest
import json

import feed
import entities
import util

class TestStop(unittest.TestCase):
  expect = {
    'stop_lat': '36.425288', 
    'stop_lon': '-117.133162', 
    'stop_id': 'FUR_CREEK_RES', 
    'stop_name': 'Furnace Creek Resort (Demo)'
  }
    
  def test_id(self):
    entity = entities.Stop(**self.expect)
    assert entity.id() == self.expect['stop_id']

  def test_name(self):
    entity = entities.Stop(**self.expect)    
    assert entity.name() == self.expect['stop_name']

  def test_point(self):
    entity = entities.Stop(**self.expect)    
    expect = (-117.133162, 36.425288)
    for i,j in zip(entity.point(), expect):
      self.assertAlmostEqual(i,j)

  def test_bbox(self):
    entity = entities.Stop(**self.expect)    
    expect = [-117.133162, 36.425288, -117.133162, 36.425288]
    for i,j in zip(entity.bbox(), expect):
      self.assertAlmostEqual(i,j)

  def test_json(self):
    entity = entities.Stop(**self.expect)    
    data = entity.json()
    assert data['name'] == entity.name()
    assert data['type'] == 'Feature'
    assert data['geometry']['type'] == 'Point'
    assert data['properties']
    # Round trip
    assert json.loads(json.dumps(data)) 

  def test_geometry(self):
    entity = entities.Stop(**self.expect)    
    expect = (-117.133162, 36.425288)
    geometry = entity.geometry()
    assert geometry['type'] == 'Point'
    for i,j in zip(geometry['coordinates'], expect):
      self.assertAlmostEqual(i,j)

  def test_routes(self):
    agency = util.preload_agency()
    entity = agency.stop(self.expect['stop_id'])
    routes = entity.routes()
    assert len(routes) == 1
    assert list(routes)[0].id() == 'BFC'


# -*- coding: utf-8 -*-
"""Two character ISO639-1 Language Codes.

From:
https://gist.github.com/alexex/4073388
"""

LANGUAGES = dict((
    ('aa', 'Afar'),
    ('ab', 'Abkhazian'),
    ('af', 'Afrikaans'),
    ('ak', 'Akan'),
    ('sq', 'Albanian'),
    ('am', 'Amharic'),
    ('ar', 'Arabic'),
    ('an', 'Aragonese'),
    ('hy', 'Armenian'),
    ('as', 'Assamese'),
    ('av', 'Avaric'),
    ('ae', 'Avestan'),
    ('ay', 'Aymara'),
    ('az', 'Azerbaijani'),
    ('ba', 'Bashkir'),
    ('bm', 'Bambara'),
    ('eu', 'Basque'),
    ('be', 'Belarusian'),
    ('bn', 'Bengali'),
    ('bh', 'Bihari languages'),
    ('bi', 'Bislama'),
    ('bo', 'Tibetan'),
    ('bs', 'Bosnian'),
    ('br', 'Breton'),
    ('bg', 'Bulgarian'),
    ('my', 'Burmese'),
    ('ca', 'Catalan; Valencian'),
    ('cs', 'Czech'),
    ('ch', 'Chamorro'),
    ('ce', 'Chechen'),
    ('zh', 'Chinese'),
    ('cu', 'Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic'),
    ('cv', 'Chuvash'),
    ('kw', 'Cornish'),
    ('co', 'Corsican'),
    ('cr', 'Cree'),
    ('cy', 'Welsh'),
    ('cs', 'Czech'),
    ('da', 'Danish'),
    ('de', 'German'),
    ('dv', 'Divehi; Dhivehi; Maldivian'),
    ('nl', 'Dutch; Flemish'),
    ('dz', 'Dzongkha'),
    ('el', 'Greek, Modern (1453-)'),
    ('en', 'English'),
    ('eo', 'Esperanto'),
    ('et', 'Estonian'),
    ('eu', 'Basque'),
    ('ee', 'Ewe'),
    ('fo', 'Faroese'),
    ('fa', 'Persian'),
    ('fj', 'Fijian'),
    ('fi', 'Finnish'),
    ('fr', 'French'),
    ('fr', 'French'),
    ('fy', 'Western Frisian'),
    ('ff', 'Fulah'),
    ('Ga', 'Georgian'),
    ('de', 'German'),
    ('gd', 'Gaelic; Scottish Gaelic'),
    ('ga', 'Irish'),
    ('gl', 'Galician'),
    ('gv', 'Manx'),
    ('el', 'Greek, Modern (1453-)'),
    ('gn', 'Guarani'),
    ('gu', 'Gujarati'),
    ('ht', 'Haitian; Haitian Creole'),
    ('ha', 'Hausa'),
    ('he', 'Hebrew'),
    ('hz', 'Herero'),
    ('hi', 'Hindi'),
    ('ho', 'Hiri Motu'),
    ('hr', 'Croatian'),
    ('hu', 'Hungarian'),
    ('hy', 'Armenian'),
    ('ig', 'Igbo'),
    ('is', 'Icelandic'),
    ('io', 'Ido'),
    ('ii', 'Sichuan Yi; Nuosu'),
    ('iu', 'Inuktitut'),
    ('ie', 'Interlingue; Occidental'),
    ('ia', 'Interlingua (International Auxiliary Language Association)'),
    ('id', 'Indonesian'),
    ('ik', 'Inupiaq'),
    ('is', 'Icelandic'),
    ('it', 'Italian'),
    ('jv', 'Javanese'),
    ('ja', 'Japanese'),
    ('kl', 'Kalaallisut; Greenlandic'),
    ('kn', 'Kannada'),
    ('ks', 'Kashmiri'),
    ('ka', 'Georgian'),
    ('kr', 'Kanuri'),
    ('kk', 'Kazakh'),
    ('km', 'Central Khmer'),
    ('ki', 'Kikuyu; Gikuyu'),
    ('rw', 'Kinyarwanda'),
    ('ky', 'Kirghiz; Kyrgyz'),
    ('kv', 'Komi'),
    ('kg', 'Kongo'),
    ('ko', 'Korean'),
    ('kj', 'Kuanyama; Kwanyama'),
    ('ku', 'Kurdish'),
    ('lo', 'Lao'),
    ('la', 'Latin'),
    ('lv', 'Latvian'),
    ('li', 'Limburgan; Limburger; Limburgish'),
    ('ln', 'Lingala'),
    ('lt', 'Lithuanian'),
    ('lb', 'Luxembourgish; Letzeburgesch'),
    ('lu', 'Luba-Katanga'),
    ('lg', 'Ganda'),
    ('mk', 'Macedonian'),
    ('mh', 'Marshallese'),
    ('ml', 'Malayalam'),
    ('mi', 'Maori'),
    ('mr', 'Marathi'),
    ('ms', 'Malay'),
    ('Mi', 'Micmac'),
    ('mk', 'Macedonian'),
    ('mg', 'Malagasy'),
    ('mt', 'Maltese'),
    ('mn', 'Mongolian'),
    ('mi', 'Maori'),
    ('ms', 'Malay'),
    ('my', 'Burmese'),
    ('na', 'Nauru'),
    ('nv', 'Navajo; Navaho'),
    ('nr', 'Ndebele, South; South Ndebele'),
    ('nd', 'Ndebele, North; North Ndebele'),
    ('ng', 'Ndonga'),
    ('ne', 'Nepali'),
    ('nl', 'Dutch; Flemish'),
    ('nn', 'Norwegian Nynorsk; Nynorsk, Norwegian'),
    ('nb', 'Bokmål, Norwegian; Norwegian Bokmål'),
    ('no', 'Norwegian'),
    ('oc', 'Occitan (post 1500)'),
    ('oj', 'Ojibwa'),
    ('or', 'Oriya'),
    ('om', 'Oromo'),
    ('os', 'Ossetian; Ossetic'),
    ('pa', 'Panjabi; Punjabi'),
    ('fa', 'Persian'),
    ('pi', 'Pali'),
    ('pl', 'Polish'),
    ('pt', 'Portuguese'),
    ('ps', 'Pushto; Pashto'),
    ('qu', 'Quechua'),
    ('rm', 'Romansh'),
    ('ro', 'Romanian; Moldavian; Moldovan'),
    ('ro', 'Romanian; Moldavian; Moldovan'),
    ('rn', 'Rundi'),
    ('ru', 'Russian'),
    ('sg', 'Sango'),
    ('sa', 'Sanskrit'),
    ('si', 'Sinhala; Sinhalese'),
    ('sk', 'Slovak'),
    ('sk', 'Slovak'),
    ('sl', 'Slovenian'),
    ('se', 'Northern Sami'),
    ('sm', 'Samoan'),
    ('sn', 'Shona'),
    ('sd', 'Sindhi'),
    ('so', 'Somali'),
    ('st', 'Sotho, Southern'),
    ('es', 'Spanish; Castilian'),
    ('sq', 'Albanian'),
    ('sc', 'Sardinian'),
    ('sr', 'Serbian'),
    ('ss', 'Swati'),
    ('su', 'Sundanese'),
    ('sw', 'Swahili'),
    ('sv', 'Swedish'),
    ('ty', 'Tahitian'),
    ('ta', 'Tamil'),
    ('tt', 'Tatar'),
    ('te', 'Telugu'),
    ('tg', 'Tajik'),
    ('tl', 'Tagalog'),
    ('th', 'Thai'),
    ('bo', 'Tibetan'),
    ('ti', 'Tigrinya'),
    ('to', 'Tonga (Tonga Islands)'),
    ('tn', 'Tswana'),
    ('ts', 'Tsonga'),
    ('tk', 'Turkmen'),
    ('tr', 'Turkish'),
    ('tw', 'Twi'),
    ('ug', 'Uighur; Uyghur'),
    ('uk', 'Ukrainian'),
    ('ur', 'Urdu'),
    ('uz', 'Uzbek'),
    ('ve', 'Venda'),
    ('vi', 'Vietnamese'),
    ('vo', 'Volapük'),
    ('cy', 'Welsh'),
    ('wa', 'Walloon'),
    ('wo', 'Wolof'),
    ('xh', 'Xhosa'),
    ('yi', 'Yiddish'),
    ('yo', 'Yoruba'),
    ('za', 'Zhuang; Chuang'),
    ('zh', 'Chinese'),
    ('zu', 'Zulu')
))

def get_language(code):
  return LANGUAGES[code.lower()]



import unittest
import json
import copy

import feed
import entities
import util

class TestRoute(unittest.TestCase):  
  expect = {
    'route_long_name': 'Airport - Bullfrog', 
    'route_id': 'AB', 
    'route_type': '3', 
    'route_text_color': '', 
    'agency_id': 'DTA', 
    'route_color': '', 
    'route_url': '', 
    'route_desc': '', 
    'route_short_name': '10'
  }

  def test_name(self):
    # Test name
    entity = entities.Route(**self.expect)    
    assert entity.name() == self.expect['route_short_name']
    # Test fallback
    expect = copy.copy(self.expect)
    expect.pop('route_short_name')
    entity = entities.Route(**expect)    
    assert entity.name() == self.expect['route_long_name']
    
  def test_id(self):
    entity = entities.Route(**self.expect)
    assert entity.id() == self.expect['route_id']
  
  # Requires preload
  def test_bbox(self):
    agency = util.preload_agency()
    route = agency.route(self.expect['route_id'])
    expect = [-116.81797, 36.868446, -116.784582, 36.88108]
    for i,j in zip(route.bbox(), expect):
      self.assertAlmostEqual(i,j)
  
  def test_json(self):
    agency = util.preload_agency()
    entity = agency.route(self.expect['route_id'])
    data = entity.json()
    assert data['name'] == entity.name()
    assert data['type'] == 'Feature'
    assert data['geometry']['type'] == 'MultiLineString'
    # Round trip
    assert json.loads(json.dumps(data)) 

  def test_geometry(self):
    agency = util.preload_agency()
    route = agency.route(self.expect['route_id'])
    geometry = route.geometry()
    assert geometry['type'] == 'MultiLineString'
    # Line 1
    expect = [
      (-116.784582, 36.868446), 
      (-116.81797, 36.88108)]
    for i,j in zip(geometry['coordinates'][0], expect):
      self.assertAlmostEqual(i[0],j[0])
      self.assertAlmostEqual(i[1],j[1])
    # Line 2
    expect = [
      (-116.81797, 36.88108), 
      (-116.784582, 36.868446)
    ]
    for i,j in zip(geometry['coordinates'][1], expect):
      self.assertAlmostEqual(i[0],j[0])
      self.assertAlmostEqual(i[1],j[1])
    
  def test_trips(self):
    agency = util.preload_agency()
    route = agency.route(self.expect['route_id'])
    assert len(route.trips()) == 2
    
  def test_stops(self):
    agency = util.preload_agency()
    route = agency.route(self.expect['route_id'])
    assert len(route.stops()) == 2


"""GTFS Agency entity."""
import entity
import geom
import util
import validation

class Agency(entity.Entity):
  """GTFS Agency entity."""
  ENTITY_TYPE = 'o'
  KEY = 'agency_id'
  REQUIRED = [
    'agency_name',
    'agency_url',
    'agency_timezone'
  ]
  OPTIONAL = [
    'agency_id',
    'agency_lang',
    'agency_phone',
    'agency_fare_url'
  ]  
  def name(self):
    return self.get('agency_name')

  def id(self):
    # agency_id is optional.
    return self.get('agency_id') or self.get('agency_name')

  def point(self):
    bbox = self.bbox()
    return [
      (bbox[0]+bbox[2])/2,
      (bbox[1]+bbox[3])/2      
    ]

  def bbox(self):
    return geom.bbox(self.stops())

  def json(self):
    return {
      'type': 'FeatureCollection',
      'name': self.name(),
      'properties': self.data,
      'bbox': self.bbox(),
      'geometry': self.geometry(),
      'routes': [r.json() for r in self.routes()],
      'features': [s.json() for s in self.stops()]
    }

  def geometry(self):
    hull = geom.convex_hull(self.stops())
    return {
      'type': 'Polygon',
      'coordinates': [
        hull + [hull[0]]
      ]
    }    

  # Agency specific methods.
  def dates(self):
    found = set(trip.get('service_id') for trip in self.trips())
    dates = [
      date 
      for date in self._feed.read('calendar') 
      if date.get('service_id') in found
    ]
    return [
      min(date.start() for date in dates),
      max(date.end() for date in dates)
    ]
  
  # Graph.      
  def routes(self):
    """Return all routes for this agency."""
    return set(self.children()) # copy
  
  def route(self, key):
    """Return a single route by ID."""
    return util.filtfirst(self.routes(), id=key)
  
  def trips(self):
    """Return all trips for this agency."""
    trips = set()
    for route in self.routes():
      trips |= route.trips()
    return trips
  
  def trip(self, key):
    """Return a single trip by ID."""
    return util.filtfirst(self.trips(), id=key)

  def stops(self):
    """Return all stops visited by trips for this agency."""
    stops = set()
    for stop_time in self.stop_times():
      stops |= stop_time.stops()
    return stops

  def stop(self, key):
    """Return a single stop by ID."""
    return util.filtfirst(self.stops(), id=key)
    
  def stop_times(self):
    """Return all stop_times for this agency."""
    stop_times = set()
    for trip in self.trips():
      stop_times |= trip.stop_times()
    return stop_times
    
  ##### Validation #####
    
  def validate(self, validator=None):
    validator = super(Agency, self).validate(validator)
    # Required
    with validator(self): 
      assert self.get('agency_name'), \
        "Required: agency_name"
    with validator(self): 
      assert self.get('agency_url'), "Required: agency_url"
      assert validation.valid_url(self.get('agency_url')), \
        "Invalid agency_url"
    with validator(self):
      assert self.get('agency_timezone'), "Required: agency_timezone"
      assert validation.valid_tz(self.get('agency_timezone')), \
        "Invalid agency_timezone"
    # Optional
    with validator(self): 
      if self.get('agency_lang'):
        assert validation.valid_language(self.get('agency_lang')), \
          "Invalid language"
    with validator(self): 
      if self.get('agency_fare_url'):
        assert validation.valid_url(self.get('agency_fare_url')), \
          "Invalid agency_fare_url"
    with validator(self):
      if self.get('agency_id'): 
        pass
    with validator(self):
      if self.get('agency_phone'):
        pass
    return validator



"""WideTime, a time class that supports long days."""
import time

class WideTime(object):
  def __init__(self, hours=0, minutes=0, seconds=0):
    assert 0 <= hours
    assert 0 <= minutes <= 60
    assert 0 <= seconds <= 60
    self.hours = hours
    self.minutes = minutes
    self.seconds = seconds
  
  def as_seconds(self):
    return self.hours * 3600 + self.minutes * 60 + self.seconds
    
  def __add__(self, other):
    return self.as_seconds() + other.as_seconds()
  
  def __sub__(self, other):
    return self.as_seconds() - other.as_seconds()
  
  @classmethod
  def from_string(cls, value):
    return cls(*map(int, value.split(':')))

  def __str__(self):
    return ':'.join('%02d'%i for i in list(self))

  def __iter__(self):
    return iter([self.hours, self.minutes, self.seconds])

  def __lt__(self, other):
    return list(self) < list(other)

  def __le__(self, other):
    return list(self) <= list(other)
  
  def __gt__(self, other):
    return list(self) > list(other)
  
  def __ge__(self, other):
    return list(self) >= list(other)
    
  def __eq__(self, other):
    return list(self) == list(other)


"""GTFS entities.

These are now split into separate modules; this module
maintains backwards compatibility.

Classes:
  Entity - GTFS entity base class
  Agency
  Route
  Trip
  StopTime
  Stop

"""
from entity import Entity
from agency import Agency
from route import Route
from trip import Trip
from stop import Stop
from stoptime import StopTime
from shape import ShapeRow, ShapeLine
from serviceperiod import ServicePeriod, ServiceDate
from farerule import FareRule
from transfer import Transfer
from frequency import Frequency
from fareattribute import FareAttribute
# from feedinfo import FeedInfo
FeedInfo = Entity



import unittest
import collections
import json

import feed
import entities
import util

class TestAgency(unittest.TestCase):
  """Test Agency."""
  expect = collections.OrderedDict({
    'agency_url': 'http://google.com', 
    'agency_name': 'Demo Transit Authority', 
    'agency_id': 'DTA', 
    'agency_timezone': 'America/Los_Angeles'
  })

  def test_name(self):
    agency = entities.Agency(**self.expect)    
    assert agency.name() == self.expect['agency_name']
    
  def test_id(self):
    agency = entities.Agency(**self.expect)    
    assert agency.id() == self.expect['agency_id']
    
  def test_feedid(self):    
    agency = entities.Agency(**self.expect)    
    assert agency.feedid('test') == 'gtfs://test/o/DTA'

  # Must get from Feed for remaining tests...
  def test_point(self):
    agency = util.preload_agency()
    point = agency.point()
    expect = [-116.76705100000001, 36.670485]
    for i,j in zip(point, expect):
      self.assertAlmostEqual(i,j)
    self.assertAlmostEqual(point[0], expect[0])
    self.assertAlmostEqual(point[1], expect[1])
    
  def test_bbox(self):
    agency = util.preload_agency()
    expect = [-117.133162, 36.425288, -116.40094, 36.915682]
    for i,j in zip(agency.bbox(), expect):
      self.assertAlmostEqual(i,j)
  
  def test_geometry(self):
    agency = util.preload_agency()
    geometry = agency.geometry()
    # a beast...
    expect = [
      (-117.133162, 36.425288), 
      (-116.40094, 36.641496), 
      (-116.751677, 36.915682), 
      (-116.76821, 36.914893), 
      (-116.81797, 36.88108),
      (-117.133162, 36.425288)
    ]
    assert geometry['type'] == 'Polygon'
    for i,j in zip(geometry['coordinates'][0], expect):
      self.assertAlmostEqual(i[0],j[0])
      self.assertAlmostEqual(i[1],j[1])
  
  def test_json(self):
    # Basic checks for JSON sanity.
    agency = util.preload_agency()
    data = agency.json()
    assert data['name'] == 'Demo Transit Authority'
    assert 'geometry' in data
    assert 'bbox' in data
    assert len(data['features']) == 9
    # Round trip
    assert json.loads(json.dumps(data)) 
  
  def test_routes(self):
    agency = util.preload_agency()
    assert len(agency.routes()) == 5
  
  def test_route(self):
    agency = util.preload_agency()
    assert agency.route('AB')
    
  def test_trips(self):
    agency = util.preload_agency()
    assert len(agency.trips()) == 11
  
  def test_trip(self):
    agency = util.preload_agency()
    assert agency.trip('AB1')
    
  def test_stops(self):
    agency = util.preload_agency()
    assert len(agency.stops()) == 9
    
  def test_stop(self):
    agency = util.preload_agency()
    assert agency.stop('NANAA')
    
  def test_stop_times(self):
    agency = util.preload_agency()
    assert len(agency.stop_times()) == 28
    


"""Validation exceptions and managers."""
import pytz
import traceback
import contextlib
import datetime

import iso639
import widetime

def make_validator(validator=None):
  return validator or ValidationManager()

def valid_color(color):
  color = str(color).lower().strip()
  if len(color) != 6:
    return False
  for j in [color[i:i+2] for i in range(0,5,2)]:
    try:
      j = int(j,16)
    except ValueError, e:
      return False
  return True
  
def valid_url(url):
  return url.startswith('http')

def valid_tz(tz):
  return pytz.timezone(tz)

def valid_language(lang):
  return iso639.get_language(lang)

def valid_int(value, vmin=None, vmax=None, empty=False):
  # Allow empty string if empty=True
  if value == '' and empty:
    return True
  try:
    value = int(value)
  except ValueError, e:
    return False
  if vmin is not None and value < vmin:
    return False    
  if vmax is not None and value > vmax:
    return False
  return True
  
def valid_float(value, vmin=None, vmax=None, empty=False):
  # Allow empty string if empty=True
  if value == '' and empty:
    return True
  try:
    value = float(value)
  except ValueError, e:
    return False
  if vmin is not None and value < vmin:
    return False    
  if vmax is not None and value > vmax:
    return False
  return True

def valid_in(value, within):
  return value in within

def valid_bool(value, empty=False):
  if value == '' and empty:
    return True
  try:
    value = int(value)
  except ValueError, e:
    return False
  return value in [0,1]

def valid_date(value, empty=False):
  if not value and empty:
    return True
  if len(value) != 8:
    return False
  try:
    datetime.datetime.strptime(value, '%Y%m%d')
  except ValueError, e:
    return False
  return True

def valid_widetime(value):
  if len(value.split(':')) != 3:
    return False
  if len(value) < 7:
    return False
  if len(value) > 8:
    return False
  try:
    widetime.WideTime.from_string(value)
  except ValueError:
    return False
  return True

def valid_point(point):
  if len(point) != 2:
    return False  
  lon, lat = point
  if not (-180 <= lon <= 180):
    return False
  if not (-90 <= lat <= 90):
    return False
  return True

##### Validation Exceptions #####

class ValidationException(Exception):
  def __init__(self, message, source=None):
    super(ValidationException, self).__init__(message)
    self.source = source

class ValidationError(ValidationException):
  """Base validation error."""
  pass

class ValidationWarning(ValidationException):
  """Validation warning."""
  pass
  
class ValidationInfo(ValidationException):
  """Validation info."""
  pass

##### Validation Managers #####

class ValidationManager(object):
  def __init__(self):
    # List of exceptions generated
    self.exceptions = []
    # Source of ValidationError
    self.source = None 
  
  def __call__(self, source=None):
    self.source = source
    return self
  
  def __enter__(self):
    return self
  
  def __exit__(self, etype, value, traceback):
    self.source = None
    return
      
  def report(self):
    print "Validation report:"
    for e in self.exceptions:
      print "%s: %s"%(e.source, e.message)

class ValidationReport(ValidationManager):
  def __exit__(self, etype, value, traceback):
    # Unset the source from this context.
    s = self.source
    self.source = None
    if not etype:
      return
    if issubclass(etype, AssertionError):
      etype = ValidationError
    if issubclass(etype, ValidationException):
      self.exceptions.append(etype(value, source=s))
      return True




"""Validate a GTFS file."""
import argparse
import json

import feed
import validation

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description='Validate a GTFS feed.')
  parser.add_argument('filename', help='GTFS File')
  parser.add_argument('--debug', 
    help='Show helpful debugging information', 
    action='store_true')
  args = parser.parse_args()
  
  validator = validation.ValidationReport()
  feed = feed.Feed(args.filename, debug=args.debug)
  feed.validate(validator=validator)
  validator.report()


"""GTFS Fare Rule."""
import datetime

import entity
import geom
import util
import widetime
import validation

class FareRule(entity.Entity):
  KEY = 'fare_id'
  REQUIRED = [
    'fare_id'
  ]
  OPTIONAL = [
    'route_id',
    'origin_id',
    'destination_id',
    'contains_id'
  ]
  
  def validate(self, validator=None):
    validator = super(FareRule, self).validate(validator)
    # Required
    with validator(self):
      assert self.get('fare_id'), "Required: fare_id"
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(FareRule, self).validate_feed(validator)
    with validator(self):
      assert self._feed.fare(self.get('fare_id')), "Unknown fare_id"
    with validator(self):
      if self.get('route_id'):
        assert self._feed.route(self.get('route_id')), "Unknown route_id"
    with validator(self):
      if self.get('origin_id'):
        pass
    with validator(self):
      if self.get('destination_id'):
        pass
    with validator(self):
      if self.get('contains_id'):
        pass
    return validator


"""GTFS StopTime entity."""
import entity
import widetime
import validation

class StopTime(entity.Entity):
  """GTFS Stop Time Entity."""
  REQUIRED = [
    'trip_id',
    'arrival_time',
    'departure_time',
    'stop_id',
    'stop_sequence',    
  ]
  OPTIONAL = [
    'stop_headsign',
    'pickup_type',
    'drop_off_type',
    'shape_dist_traveled',
    'timepoint'
  ]
  
  def point(self):
    # Ugly hack.
    return list(self._children)[0].point()

  def arrive(self):
    if self.get('arrival_time'):
      return widetime.WideTime.from_string(self.get('arrival_time'))
    
  def depart(self):
    if self.get('departure_time'):
      return widetime.WideTime.from_string(self.get('departure_time'))
    
  def sequence(self):
    return int(self.get('stop_sequence'))  
    
  # Graph
  def stops(self):
    return set(self.children())

  ##### Validation #####
  def validate(self, validator=None):
    validator = super(StopTime, self).validate(validator)
    # Required
    with validator(self):
      assert self.get('trip_id'), "Required: trip_id"
    with validator(self):
      assert \
        (self.get('arrival_time') and self.get('departure_time')) or \
        (not self.get('arrival_time') and not self.get('departure_time')), \
        "Both arrival_time and departure_time must be set, or both must be empty."
    with validator(self):
      if self.get('arrival_time'):
        assert validation.valid_widetime(self.get('arrival_time')), \
          "Invalid arrival_time: %s"%self.get('arrival_time')
    with validator(self):
      if self.get('departure_time'):
        assert validation.valid_widetime(self.get('departure_time')), \
          "Invalid departure_time: %s"%self.get('departure_time')
    with validator(self): 
      assert self.arrive() <= self.depart(), \
        "Cannot depart before arriving!: %s -> %s"%(self.arrive(), self.depart())
    with validator(self):
      assert self.sequence() >= 0, \
        "Invalid stop_sequence: %s"%self.get('stop_sequence')
    # TODO: Warnings - useless stops (cant pickup or dropoff)
    # Optional
    with validator(self):
      if self.get('pickup_type'):
        assert validation.valid_int(self.get('pickup_type'), vmin=0, vmax=3), \
          "Invalid pickup_type, must be 0,1,2,3: %s"%self.get('pickup_type')
    with validator(self): 
      if self.get('drop_off_type'):
        assert validation.valid_int(self.get('drop_off_type'), vmin=0, vmax=3), \
          "Invalid drop_off_type, must be 0,1,2,3: %s"%self.get('drop_off_type')
    with validator(self):
      if self.get('timepoint'):
        assert validation.valid_bool(self.get('timepoint'), empty=True), \
          "Invalid timepoint"
    with validator(self):
      if int(self.get('timepoint',0)) == 1:
        assert self.arrive() and self.depart(), \
          "Exact timepoints require arrival_time and departure_time"
    with validator(self):
      if self.get('stop_headsign'):
        pass
    with validator(self):
      if self.get('shape_dist_traveled'):
        pass
    return validator

  def validate_feed(self, validator=None):
    # TODO: Check this in feed, since it can be a faster set operation?
    validator = super(StopTime, self).validate_feed(validator)
    with validator(self):
      assert self._feed.trip(self.get('trip_id')), "Unknown trip_id"
    with validator(self):
      assert self._feed.stop(self.get('stop_id')), "Unknown stop_id"
    return validator
    



"""Provide useful information about a GTFS file and export to JSON."""
import argparse
import json

import feed

if __name__ == "__main__":
  parser = argparse.ArgumentParser(description='GTFS Info and JSON export')
  parser.add_argument('filename', help='GTFS File')
  parser.add_argument('--debug', 
    help='Show helpful debugging information', 
    action='store_true')

  args = parser.parse_args()
  g = feed.Feed(args.filename)
  f.preload()
  print "===== GTFS: %s ====="%g.filename
  for agency in g.agencies():
    print "Agency:", agency['agency_name']
    print "  Routes:", len(agency.routes())
    if args.debug:
      for route in agency.routes():
        print route.data
    print "  Stops:", len(agency.stops())
    if args.debug:
      for stop in agency.stops():
        print stop.data
    print "  Trips:", len(agency.trips())
    if args.debug:
      for trip in agency.trips():
        print trip.data

    # Export
    outfile = 'export-%s.geojson'%agency.id()
    print "Writing: %s"%outfile
    with open(outfile, 'w') as f:
      json.dump(
        agency.json(), 
        f, 
        sort_keys=True, 
        indent=4, 
        separators=(',', ': ')
      )



"""Base entity."""

import validation

class Entity(object):
  """A GTFS Entity.
  
  Parent-child relationships:
  
    Agency -> 
      Route -> 
        Trip -> 
          StopTime -> 
            Stop
  
  """
  ENTITY_TYPE = None
  KEY = None
  REQUIRED = []
  OPTIONAL = []
  
  def __init__(self, **data):
    """Row data from DictReader, and reference to feed."""
    # The row data. 
    # This is a collections.namedtuple, and is read-only.
    self.data = data
    # Reference to GTFS Reader.
    self._feed = None
    # Relationships (e.g. trips, stop_times, ...)
    self._children = set()  
    self._parents = set()
  
  # GTFS row data.
  def __len__(self):
    return len(self.data)
  
  def __contains__(self, key):
    return key in self.data
  
  def __getitem__(self, key):
    """Proxy to row data."""
    # Work with either a dict or a namedtuple.
    if hasattr(self.data, '_fields'):
      try:
        return getattr(self.data, key)
      except AttributeError:
        raise KeyError
    else:
      return self.data[key]

  def get(self, key, default=None):
    """Get row data by key."""
    try:
      return self[key]
    except KeyError:
      return default
      
  def set(self, key, value):
    # Convert from namedtuple to dict if setting value.
    if hasattr(self.data, '_fields'):
      self.data = self.data._asdict()
    self.data[key] = value
  
  def set_feed(self, feed):
    self._feed = feed
  
  def keys(self):
    if hasattr(self.data, '_fields'):
      return self.data._fields
    return self.data.keys()
    
  def items(self):
    if hasattr(self.data, '_fields'):
      return self.data._asdict().items()
    return self.data.items()
  
  # Name methods.
  def name(self):
    """A reasonable name for the entity."""
    raise NotImplementedError
  
  def id(self):
    """An internal GTFS identifier, e.g. route_id."""
    raise NotImplementedError
  
  def feedid(self, feedid):
    """A canonical Onestop-style ID for this entity."""
    return 'gtfs://%s/%s/%s'%(
      feedid,
      self.ENTITY_TYPE,
      self.id()
    )

  # Entity geometry.
  def point(self):
    """Return a point for this entity."""
    raise NotImplementedError  
    
  def bbox(self):
    """Return a bounding box for this entity."""
    raise NotImplementedError
  
  def geometry(self):
    """Return a GeoJSON-type geometry for this entity."""
    raise NotImplementedError
    
  # Load / Dump.
  @classmethod
  def from_row(cls, data, feed=None):
    entity = cls()
    entity.data = data
    entity.set_feed(feed)
    return entity
    
  @classmethod
  def from_json(cls, data):
    raise NotImplementedError
  
  def json(self):
    raise NotImplementedError

  # Graph.
  def pclink(self, parent, child):
    """Create a parent-child relationship."""
    if parent._children is None:
      parent._children = set()
    if child._parents is None:
      child._parents = set()
    parent._children.add(child)
    child._parents.add(parent)
    
  # ... children
  def add_child(self, child):
    """Add a child relationship."""
    self.pclink(self, child)
    
  def children(self):
    """Read and cache children."""
    return self._children
    
  # ... parents
  def add_parent(self, parent):
    """Add a parent relationship."""
    self.pclink(parent, self)
    
  def parents(self):
    """Read and cache parents."""
    return self._parents
    
  def _read_parents(self):
    """Read the parents from the GTFS feed."""
    return set()
    
  ##### Validation #####
  
  def validate(self, validator=None):
    return validation.make_validator(validator)
    
  def validate_feed(self, validator=None):
    return validation.make_validator(validator)
    
    


"""GTFS ShapeRow entity; these can be collected into a ShapeLine"""
import entity
import validation

class ShapeLine(entity.Entity):
  """A collection of ShapeRows."""
  def rows(self):
    return sorted(
      self._children, 
      key=lambda x:int(x.get('shape_pt_sequence',0))
    )

  def points(self):
    return [i.point() for i in self.rows()]

  def geometry(self):
    return {
      'type':'MultiLineString',
      'coordinates': [self.points()]
    }
  
  def json(self):
    return {
      'type': 'Feature',
      'properties': self.data,
      'geometry': self.geometry()
    }
      
class ShapeRow(entity.Entity):
  """A row in shapes.txt"""
  REQUIRED = [
    'shape_id',
    'shape_pt_lat',
    'shape_pt_lon',
    'shape_pt_sequence',
    
  ]
  OPTIONAL = [
    'shape_dist_traveled',
  ]
  
  def id(self):
    return self.get('shape_id')

  def name(self):
    return self.get('shape_id')  
  
  def point(self):
    try:
      return float(self.get('shape_pt_lon')), float(self.get('shape_pt_lat'))
    except ValueError, e:
      return None

  def geometry(self):
    return {
      "type": 'Point',
      "coordinates": self.point(),
    }

  def validate(self, validator=None):
    validator = super(ShapeRow, self).validate(validator)
    # Required
    with validator(self):
      assert validation.valid_int(self.get('shape_pt_sequence'), vmin=0), \
        "Required: shape_pt_sequence"
    with validator(self): 
      assert self.get('shape_pt_lat'), "Required: shape_pt_lat"
    with validator(self): 
      assert self.get('shape_pt_lon'), "Required: shape_pt_lon"
    with validator(self): 
      assert validation.valid_point(self.point()), \
      "Invalid shape_pt_lon or shape_pt_lat"
    # Optional
    return validator



"""Feed unit tests."""
import unittest
import os
import json
import inspect
import tempfile
import csv
import zipfile

import util
import feed
import entities

def test_outfile():
  # Create a temporary filename.
  # You'll have to unlink file when done.
  # This of course is not secure against file creation problems.
  outfile = tempfile.NamedTemporaryFile()
  name = outfile.name
  outfile.close()
  return name

class TestFeed(unittest.TestCase):
  """Test Feed Reader.
  
  TODO: Test Unicode?
  
  """  
  agency_expect = {
    'agency_url': 'http://google.com', 
    'agency_name': 'Demo Transit Authority', 
    'agency_id': 'DTA', 
    'agency_timezone': 'America/Los_Angeles'
  }
  
  route_expect = {
    'route_long_name': 'Airport - Bullfrog', 
    'route_id': 'AB', 
    'route_type': '3', 
    'route_text_color': '', 
    'agency_id': 'DTA', 
    'route_color': '', 
    'route_url': '', 
    'route_desc': '', 
    'route_short_name': '10'
  }
  
  stop_expect = {
    'stop_lat': '36.425288', 
    'stop_lon': '-117.133162', 
    'stop_id': 'FUR_CREEK_RES', 
    'stop_name': 'Furnace Creek Resort (Demo)'
  }

  def test_init(self):
    f = feed.Feed(util.example_feed())
  
  def test_read(self):
    # Test basic read
    f = feed.Feed(util.example_feed())
    data = f.read('stops')
    # check we got 9 entities
    assert len(data) == 9
    # check cache
    assert 'stops' in f.by_id
  
  def test_read_path(self):
    # Test overlay
    f = feed.Feed(
      util.example_feed(), 
      path=os.path.dirname(util.example_feed())
    )
    assert f.stop('TEST')
    with self.assertRaises(Exception):
      f.stop('FUR_CREEK_RES')
  
  def test_read_missing(self):
    f = feed.Feed(
      util.example_feed(), 
      path=os.path.dirname(util.example_feed())
    )
    with self.assertRaises(Exception):
      f.read('missing')
  
  def test_write(self):
    f = feed.Feed()
    data = [entities.Agency(**self.agency_expect)]
    outfile = test_outfile()
    f.write(outfile, data, sortkey='agency_id')
    # Check the output...
    with open(outfile) as csvfile:
      reader = csv.reader(csvfile)
      headers = reader.next()
      assert len(self.agency_expect.keys()) == len(headers)
      for i in headers:
        assert i in self.agency_expect
      rows = []
      for i in reader:
        rows.append(i)
      assert len(rows) == 1
      row = rows[0]
      for k,v in zip(headers, row):
        assert self.agency_expect[k] == v
    # Delete temp file
    os.unlink(outfile)

  def test_write_exists(self):
    f = feed.Feed()
    data = [entities.Agency(**self.agency_expect)]
    outfile = test_outfile()
    f.write(outfile, data, sortkey='agency_id')
    with self.assertRaises(IOError):
      f.write(outfile, data, sortkey='agency_id')
    os.unlink(outfile)

  def test_make_zip(self):
    f = feed.Feed()
    outfile = test_outfile()
    f.make_zip(
      outfile,
      path=os.path.dirname(util.example_feed()),
      clone=util.example_feed()
    )
    expect = [
      'agency.txt', 
      'calendar.txt', 
      'calendar_dates.txt', 
      'fare_attributes.txt', 
      'fare_rules.txt', 
      'frequencies.txt', 
      'routes.txt', 
      'shapes.txt', 
      'stop_times.txt', 
      'trips.txt', 
      'stops.txt'
    ]
    zf = zipfile.ZipFile(outfile)
    for i,j in zip(sorted(zf.namelist()), sorted(expect)):
      assert i == j
    zf.close()
    os.unlink(outfile)

  def test_make_zip_exists(self):
    f = feed.Feed()
    outfile = test_outfile()
    f.make_zip(
      outfile,
      path=os.path.dirname(util.example_feed()),
      clone=util.example_feed()
    )
    with self.assertRaises(IOError):
      f.make_zip(
        outfile,
        path=os.path.dirname(util.example_feed()),
        clone=util.example_feed()
      )
    os.unlink(outfile)

  def test_cache(self):
    f = feed.Feed(util.example_feed())
    # Read a first time
    data1 = f.read('stops')
    # Read a second time
    data2 = f.read('stops')
    assert len(data1) == len(data2)
    assert 'stops' in f.by_id
    assert len(data1) == len(f.by_id['stops'])

  def test_read_invalidfile(self):
    f = feed.Feed(util.example_feed())
    with self.assertRaises(KeyError):
      f.read('invalidfile')
      
  def test_read_padding(self):
    # The Google GTFS example feed is missing columns in
    # stop_times.txt. Check the padding mechanism works.
    f = feed.Feed(util.example_feed())
    data = f.read('stop_times')
    # Check that all 9 elements are present.
    for entity in f.read('stop_times'):
      assert len(entity) == 9

  def test_agencies(self):
    f = feed.Feed(util.example_feed())
    data = f.agencies()
    assert len(data) == 1
    
  def test_agency(self):
    f = feed.Feed(util.example_feed())
    data = f.agency(self.agency_expect['agency_id'])
    for k in self.agency_expect:
      assert self.agency_expect[k] == data[k]
  
  def test_routes(self):
    f = feed.Feed(util.example_feed())
    assert len(f.routes()) == 5
    
  def test_route(self):
    f = feed.Feed(util.example_feed())
    data = f.route(self.route_expect['route_id'])
    for k in self.route_expect:
      assert self.route_expect[k] == data[k]

  def test_stops(self):
    f = feed.Feed(util.example_feed())
    assert len(f.stops()) == 9

  def test_stop(self):
    f = feed.Feed(util.example_feed())
    data = f.stop(self.stop_expect['stop_id'])
    for k in self.stop_expect:
      assert self.stop_expect[k] == data[k]



"""Geometry utilities."""
# Bounding box
def bbox(features):
  points = [s.point() for s in features]
  if not points:
    raise ValueError("Cannot create bbox with no features.")
  lons = [p[0] for p in points]
  lats = [p[1] for p in points]
  return [
    min(lons),
    min(lats),
    max(lons),
    max(lats)
  ]
  
# Convex hull - Graham Scan - Tom Switzer <thomas.switzer@gmail.com>
# http://tomswitzer.net/2010/03/graham-scan/
TURN_LEFT, TURN_RIGHT, TURN_NONE = (1, -1, 0)
 
def _turn(p, q, r):
    return cmp((q[0] - p[0])*(r[1] - p[1]) - (r[0] - p[0])*(q[1] - p[1]), 0)
 
def _keep_left(hull, r):
    while len(hull) > 1 and _turn(hull[-2], hull[-1], r) != TURN_LEFT:
            hull.pop()
    if not len(hull) or hull[-1] != r:
        hull.append(r)
    return hull
 
def convex_hull(features):
    """Returns points on convex hull of an array of points in CCW order."""
    points = sorted([s.point() for s in features])
    l = reduce(_keep_left, points, [])
    u = reduce(_keep_left, reversed(points), [])
    return l.extend(u[i] for i in xrange(1, len(u) - 1)) or l




"""Entity unit tests."""
import unittest
import os
import json
import collections
import copy

import feed
import entities
import util

class TestEntity(unittest.TestCase):
  """Test Entity methods, and unimplemented abstract methods."""
  expect = collections.OrderedDict({
    'foo':'bar',
    'rab':'oof'
  })

  def test_init(self):
    agency = entities.Entity(**self.expect)
  
  def test_get(self):
    agency = entities.Entity(**self.expect)
    for key in self.expect.keys():
      assert agency.get(key) == self.expect.get(key)
      assert agency[key] == self.expect[key]

  def test_get_keyerror(self):
    agency = entities.Entity(**self.expect)    
    with self.assertRaises(KeyError):
      agency['asdf']
  
  def test_get_default(self):
    agency = entities.Entity(**self.expect)    
    assert agency.get('asdf','test') == 'test'
    
  def test_get_namedtuple(self):
    nt = collections.namedtuple('test', self.expect.keys())
    agency = entities.Entity.from_row(nt(**self.expect))
    for key in self.expect.keys():
      assert agency.get(key) == self.expect.get(key)
      assert agency[key] == self.expect[key]

  def test_get_namedtuple_keyerror(self):
    nt = collections.namedtuple('test', self.expect.keys())
    agency = entities.Entity.from_row(nt(**self.expect))
    with self.assertRaises(KeyError):
      agency['asdf']

  def test_setitem(self):
    agency = entities.Entity(**self.expect)
    agency.set('test', 'ok')
    assert agency['test'] == 'ok'
  
  def test_setitem_ntconvert(self):
    nt = collections.namedtuple('test', self.expect.keys())
    agency = entities.Entity.from_row(nt(**self.expect))
    agency.set('test', 'ok')
    assert agency['test'] == 'ok'    

  def test_name(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.name()

  def test_id(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.id()

  def test_point(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.point()

  def test_bbox(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.bbox()

  def test_geometry(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.geometry()

  def test_from_json(self):
    with self.assertRaises(NotImplementedError):
      entities.Entity.from_json(self.expect)

  def test_json(self):
    entity = entities.Entity(**self.expect)
    with self.assertRaises(NotImplementedError):
      entity.json()

  # Relationships
  def test_pclink(self):
    agency1 = entities.Entity(**self.expect)    
    agency2 = entities.Entity(**self.expect)    
    agency1.pclink(agency1, agency2)
    assert agency2 in agency1.children()
    assert agency1 in agency2.parents()

  def test_add_child(self):
    agency1 = entities.Entity(**self.expect)    
    agency2 = entities.Entity(**self.expect)    
    agency1.add_child(agency2)
    assert agency2 in agency1.children()
    assert agency1 in agency2.parents()

  def test_add_parent(self):
    agency1 = entities.Entity(**self.expect)    
    agency2 = entities.Entity(**self.expect)    
    agency2.add_parent(agency1)
    assert agency2 in agency1.children()
    assert agency1 in agency2.parents()

  def test_parents(self):
    self.test_add_parent()
    entity = entities.Entity(**self.expect)    
    assert not entity.parents()
  
  def test_children(self):
    self.test_add_child()
    entity = entities.Entity(**self.expect)    
    assert not entity.children()
    
    




"""GTFS Trip entity."""
import entity
import validation

class Trip(entity.Entity):
  """GTFS Trip entity."""
  ENTITY_TYPE = 't'
  KEY = 'trip_id'
  REQUIRED = [
    'trip_id',
    'route_id',
    'service_id',
  ]
  OPTIONAL = [
    'trip_headsign',
    'trip_short_name',
    'direction_id',
    'block_id',
    'shape_id',
    'wheelchair_accessible',
    'bikes_allowed',
  ]  
  def id(self):
    return self.get('trip_id')
  
  def start(self):
    return self.stop_sequence()[0].arrive()
  
  def end(self):
    return self.stop_sequence()[-1].arrive()
  
  # Graph
  def stop_times(self):
    return set(self.children()) # copy

  # Trip methods.
  def stop_sequence(self):
    """Return the sorted StopTimes for this trip."""
    return sorted(
      self.stop_times(), 
      key=lambda x:int(x.get('stop_sequence'))
    )
    
  ##### Validation #####
  def validate(self, validator=None):
    validator = super(Trip, self).validate(validator)
    # Required
    with validator(self): 
      assert self.get('route_id'), "Required: route_id"
    with validator(self): 
      assert self.get('service_id'), "Required: service_id"
    with validator(self):
      assert self.get('trip_id'), "Required: trip_id"
    # TODO: Warnings:
    #   speed/vehicle_type
    #   duplicate trips
    # Optional
    with validator(self):
      if self.get('direction_id'):
        assert validation.valid_bool(self.get('direction_id')), "Invalid direction_id"
    with validator(self):
      if self.get('wheelchair_accessible'):
        assert validation.valid_int(self.get('wheelchair_accessible'), vmin=0, vmax=2, empty=True), \
          "Invalid wheelchair_accessible"
    with validator(self):
      if self.get('bikes_allowed'):
        assert validation.valid_int(self.get('bikes_allowed'), vmin=0, vmax=2, empty=True), \
          "Invalid bikes_allowed"
    with validator(self):
      if self.get('trip_headsign'):
        pass
    with validator(self):
      if self.get('trip_short_name'):
        pass
    with validator(self):
      if self.get('block_id'):
        pass
    with validator(self):
      if self.get('shape_id'):
        pass

  def validate_feed(self, validator=None):
    validator = super(Trip, self).validate_feed(validator)
    with validator(self):
      assert self._feed.route(self.get('route_id')), "Unknown route_id"
    with validator(self):
      assert self._feed.serviceperiod(self.get('service_id')), "Unknown service_id"
    with validator(self):
      cur = 0
      for i in self.stop_sequence():
        j = int(i.get('stop_sequence'))
        assert j > cur, \
          "Invalid stop_time sequence: stop_sequence must increase"
        cur = j
    # TODO: validate shape
    return validator
    


"""GTFS Frequency."""
import datetime

import entity
import geom
import util
import widetime
import validation

class Frequency(entity.Entity):
  REQUIRED = [
    'trip_id',
    'start_time',
    'end_time',
    'headway_secs'
  ]
  OPTIONAL = [
    'exact_times'
  ]
  
  def start(self):
    return widetime.WideTime.from_string(self.get('start_time'))
    
  def end(self):
    return widetime.WideTime.from_string(self.get('end_time'))

  def validate(self, validator=None):
    validator = super(Frequency, self).validate(validator)
    # Required
    with validator(self):
      assert self.get('trip_id'), "Required: trip_id"
    with validator(self):
      assert self.get('start_time'), "Required: start_time"
    with validator(self):
      assert self.get('end_time'), "Required: end_time"
    with validator(self):
      assert validation.valid_widetime(self.get('start_time')), \
        "Invalid start_time"
    with validator(self):
      assert validation.valid_widetime(self.get('end_time')), \
        "Invalid end_time"
    with validator(self):
      assert self.end() >= self.start(), \
        "Invalid end_time, must at least start_date"
    with validator(self):
      assert validation.valid_int(self.get('headway_secs'), vmin=1), \
        "Invalid headway_secs"
    # Optional
    with validator(self):
      if self.get('exact_times'):
        assert validation.valid_bool(self.get('exact_times'), empty=True), \
          "Invalid exact_times"
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(FareRule, self).validate_feed(validator)
    with validator(self):
      assert self._feed.trip(self.get('trip_id')), "Unknown trip_id"
    return validator


"""GTFS ServicePeriod entity."""
import datetime

import entity
import geom
import util
import validation

class ServicePeriod(entity.Entity):
  KEY = 'service_id'
  REQUIRED = [
    'service_id',
    'monday',
    'tuesday',
    'wednesday',
    'thursday',
    'friday',
    'saturday',
    'sunday',
    'start_date',
    'end_date'
  ]
  OPTIONAL = [
  ]

  def start(self):
    return datetime.datetime.strptime(self.get('start_date'), '%Y%m%d')

  def end(self):
    return datetime.datetime.strptime(self.get('end_date'), '%Y%m%d')

  def validate(self, validator=None):
    validator = super(ServicePeriod, self).validate(validator)
    with validator(self):
      assert self.get('service_id'), "Required: service_id"
    with validator(self):
      assert validation.valid_bool(self.get('monday')), "Required: monday"
    with validator(self):
      assert validation.valid_bool(self.get('tuesday')), "Required: tuesday"
    with validator(self):
      assert validation.valid_bool(self.get('wednesday')), "Required: wednesday"
    with validator(self):
      assert validation.valid_bool(self.get('thursday')), "Required: thursday"
    with validator(self):
      assert validation.valid_bool(self.get('friday')), "Required: friday"
    with validator(self):
      assert validation.valid_bool(self.get('saturday')), "Required: saturday"
    with validator(self):
      assert validation.valid_bool(self.get('sunday')), "Required: sunday"
    with validator(self):
      assert validation.valid_date(self.get('start_date'), empty=True), "Invalid start_date"
    with validator(self):
      assert validation.valid_date(self.get('start_end'), empty=True), "Invalid start_end"
    with validator(self):
      assert self.end() >= self.start(), \
        "Invalid end_date, must be at least start_date"
    # TODO: Warnings
    #   - no days of the week
    return validator
  
class ServiceDate(entity.Entity):
  REQUIRED = [
    'service_id',
    'date',
    'exception_type'
  ]
  
  def validate(self, validator=None):
    validator = super(ServiceDate, self).validate(validator)
    with validator(self):
      assert self.get('service_id'), "Required: service_id"
    with validator(self):
      assert self.get('date'), "Required: date"
    with validator(self):
      assert validation.valid_int(self.get('exception_type'), vmin=1, vmax=2), \
        "Invalid exception_type"
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(ServiceDate, self).validate_feed(validator)
    with validator(self):
      assert self._feed.serviceperiod(self.get('service_id')), \
        "Unknown service_id"    
    return validator


from setuptools import setup

import mzgtfs

setup(
  name='mzgtfs',
  version=mzgtfs.__version__,
  description='Mapzen GTFS',
  author='Ian Rees',
  author_email='ian@mapzen.com',
  url='https://github.com/transitland/mapzen-gtfs',
  license='License :: OSI Approved :: MIT License',
  packages=['mzgtfs'],
  install_requires=['unicodecsv', 'pytz'],
  zip_safe=False,
  # Include examples.
  package_data = {
    '': ['*.txt', '*.md', '*.zip']
  }
)


"""Test utilities."""
import unittest
import os
import json
import inspect

import util
import feed
import entities

class TestUtil(unittest.TestCase):
  testdata = [
    entities.Agency(agency_name='Foo', agency_id='1'),
    entities.Agency(agency_name='Bar', agency_id='2'),
    entities.Agency(agency_name='Baz', agency_id='2'),
  ]

  def test_filtany_name(self):
    data = util.filtany(self.testdata, name='Foo')
    assert len(data) == 1
    data = util.filtany(self.testdata, name='Bar')
    assert len(data) == 1
    data = util.filtany(self.testdata, name='Baz')
    assert len(data) == 1

  def test_filtany_id(self):
    data = util.filtany(self.testdata, id='1')
    assert len(data) == 1
    data = util.filtany(self.testdata, id='2')
    assert len(data) == 2
    
  def test_filtfirst_name(self):
    data = util.filtfirst(self.testdata, name='Foo')
    assert data.name() == 'Foo'

  def test_filtfirst_id(self):
    data = util.filtfirst(self.testdata, id='1')
    assert data.id() == '1'
  
  def test_filtfirst_empty(self):
    with self.assertRaises(ValueError):
      util.filtfirst(self.testdata, name='Asdf')
    with self.assertRaises(ValueError):
      util.filtfirst(self.testdata, id='0')
      
  def test_example_feed(self):
    expect = util.example_feed()
    assert os.path.exists(expect)



"""Test utilities."""
import unittest
import os
import json
import inspect

import geom

class MockPoint(object):
  def __init__(self, x, y):
    self.x = x
    self.y = y
  def point(self):
    return [self.x,self.y]

def mockpoints():
  return [
    MockPoint(0,1),
    MockPoint(1,0),
    MockPoint(1,1)
  ]

class TestGeom(unittest.TestCase):
  def test_bbox(self):
    expect = [0, 0, 1, 1]
    data = geom.bbox(mockpoints())
    for i,j in zip(data, expect):
      assert i == j
      
  def test_bbox_nopoints(self):
    with self.assertRaises(ValueError):
      geom.bbox([])
      
  def test_convex_hull(self):
    expect = [[0, 1], [1, 0], [1, 1]]
    data = geom.convex_hull(mockpoints())
    for i,j in zip(data, expect):
      self.assertAlmostEqual(i[0], j[0])
      self.assertAlmostEqual(i[1], j[1])



"""GTFS Feed Reader."""
import csv
import zipfile
import collections
import json
import os
import tempfile
import glob
import subprocess
import csv

try:
  import unicodecsv
except ImportError:
  unicodecsv = None

import util
import entities
import validation
  
class Feed(object):
  """Read a GTFS feed."""

  # Classes for each GTFS table.
  FACTORIES = {
    'agency': entities.Agency,
    'routes': entities.Route,
    'trips': entities.Trip,
    'stops': entities.Stop,
    'stop_times': entities.StopTime,
    'shapes': entities.ShapeRow,
    'calendar': entities.ServicePeriod,
    'calendar_dates': entities.ServiceDate,
    'fare_rules': entities.FareRule,
    'fare_attributes': entities.FareAttribute,
    'transfers': entities.Transfer,
    'frequencies': entities.Frequency,
    'feed_info': entities.FeedInfo
  }

  def __init__(self, filename=None, path=None, debug=False):
    """Filename required."""
    self.filename = filename
    self.path = path
    self.debug = debug
    self.cache = {}
    self.by_id = {}
    self._shapes = None
    self._zones = None

  ##### Read / write #####
  
  def log(self, msg):
    if self.debug:
      print msg

  def _open(self, table):
    arcname = '%s.txt'%table
    f = None
    zf = None
    if self.path and os.path.exists(os.path.join(self.path, arcname)):
      f = open(os.path.join(self.path, arcname))
    elif self.filename and os.path.exists(self.filename):
      zf = zipfile.ZipFile(self.filename)
      try:
        f = zf.open(arcname)
      except KeyError:
        pass
    elif self.filename and not os.path.exists(self.filename):
      raise KeyError("File not found: %s"%self.filename)
    if not f:
      raise KeyError("File not found in path or zip file: %s"%arcname)
    return f

  def iterread(self, table):
    """Iteratively read data from a GTFS table. Returns namedtuples."""
    self.log('Reading: %s'%table)
    # Entity class
    cls = self.FACTORIES[table]
    f = self._open(table)
    # csv reader
    if unicodecsv:
      data = unicodecsv.reader(f, encoding='utf-8-sig')
    else:
      data = csv.reader(f)
    header = data.next()
    headerlen = len(header)
    ent = collections.namedtuple(
      'EntityNamedTuple', 
      map(str, header)
    )
    for row in data:
      if len(row) == 0:
        continue
      # Get rid of extra spaces.
      row = [i.strip() for i in row]
      # pad to length if necessary... :(
      if len(row) < headerlen:
        row += ['']*(headerlen-len(row))
      yield cls.from_row(ent._make(row), self)
    f.close()
    
  def read(self, table):
    """..."""
    # Table exists
    if table in self.by_id:
      return self.by_id[table].values()
    if table in self.cache:
      return self.cache[table]    
    # Read table
    cls = self.FACTORIES[table]
    key = cls.KEY
    if key:
      if table not in self.by_id:
        self.by_id[table] = {}
      t = self.by_id[table]
      for item in self.iterread(table):
        t[item.get(key)] = item
      return t.values()
    if table not in self.cache:
      self.cache[table] = []
    t = self.cache[table]
    for item in self.iterread(table):
      t.append(item)
    return t

  def write(self, filename, entities, sortkey=None, columns=None):
    """Write entities out to filename in csv format.
    
    Note: this doesn't write directly into a Zip archive, because this behavior
    is difficult to achieve with Zip archives. Use make_zip() to create a new
    GTFS Zip archive.
    """    
    if os.path.exists(filename):
      raise IOError('File exists: %s'%filename)
    # Make sure we have all the entities loaded.
    if sortkey:
      entities = sorted(entities, key=lambda x:x[sortkey])
    if not columns:
      columns = set()
      for entity in entities:
        columns |= set(entity.keys())
      columns = sorted(columns)      
    # Write the csv file
    with open(filename, 'wb') as f:
      writer = unicodecsv.writer(f) # , encoding='utf-8-sig'
      writer.writerow(columns)
      for entity in entities:
        writer.writerow([entity.get(column) for column in columns])

  def make_zip(self, filename, files=None, path=None, clone=None):
    """Create a Zip archive.
    
    Provide any of the following:
      files - A list of files
      path - A directory of .txt files
      clone - Copy any files from a zip archive not specified above

    Duplicate files will be ignored. The 'files' argument will be used first,
    then files found in the specified 'path', then in the 
    specified 'clone' archive.
    """
    if filename and os.path.exists(filename):
      raise IOError('File exists: %s'%filename)     
    files = files or []
    arcnames = []
    if path and os.path.isdir(path):
      files += glob.glob(os.path.join(path, '*.txt'))
    # Write files.
    self.log("Creating zip archive: %s"%filename)
    zf = zipfile.ZipFile(filename, 'a')
    for f in files:
      base = os.path.basename(f)
      if base in arcnames:
        self.log('... skipping: %s'%f)
      else:
        self.log('... adding: %s'%f)
        arcnames.append(base)
        zf.write(f, base)

    # Clone from existing zip archive.
    if clone and os.path.exists(clone):
      zc = zipfile.ZipFile(clone)
      for f in zc.namelist():
        base = os.path.basename(f)
        if os.path.splitext(base)[-1] != '.txt':
          pass
          # self.log('... skipping from clone: %s'%f)
        elif base in arcnames:
          self.log('... skipping from clone: %s'%f)
        else:
          self.log('... adding from clone: %s'%f)
          arcnames.append(base)
          with zc.open(f) as i:
            data = i.read()
          zf.writestr(base, data)
    zf.close()

  def preload(self):
    # Load tables with primary key
    for table,cls in self.FACTORIES.items():
      if not cls.KEY:
        continue
      try:
        self.read(table)
      except KeyError:
        pass

    default_agency_id = None
    agencies = self.agencies()
    if len(agencies) == 1:
      default_agency_id = agencies[0].get('agency_id')

    for route in self.routes():
      route.add_parent(self.agency(route.get('agency_id') or default_agency_id))
    for trip in self.trips():
      trip.add_parent(self.route(trip.get('route_id')))
    # Load stop_times  
    for stoptime in self.read('stop_times'):
      stoptime.add_parent(self.trip(stoptime.get('trip_id')))
      stoptime.add_child(self.stop(stoptime.get('stop_id')))

  ##### Keyed entities #####
  
  def _entities(self, table):
    return self.read(table)
  
  def _entity(self, table, key):
    if table not in self.by_id:
      self.read(table)
    return self.by_id[table][key]
    
  def agencies(self): return self._entities('agency')
  def agency(self, key): return self._entity('agency', key)
  def routes(self): return self._entities('routes')
  def route(self, key): return self._entity('routes', key)
  def stops(self): return self._entities('stops')
  def stop(self, key): return self._entity('stops', key)
  def trips(self): return self._entities('trips')
  def trip(self, key): return self._entity('trips', key)
  def fares(self): return self._entities('fare_attributes')
  def fare(self, key): return self._entity('fare_attributes', key)
  def serviceperiods(self): return self._entities('calendar')
  def serviceperiod(self, key): return self._entity('calendar', key)

  def shapes(self):
    """Return the route shapes as a dictionary."""
    # Todo: Cache?
    if self._shapes:
      return self._shapes
    # Group together by shape_id
    self.log("Generating shapes...")
    ret = collections.defaultdict(entities.ShapeLine)
    for point in self.read('shapes'):
      ret[point['shape_id']].add_child(point)
    self._shapes = ret
    return self._shapes
    
  ##### Other methods #####
  
  def dates(self):
    data = self.read('calendar')
    return [
      min(i.start() for i in data),
      max(i.end() for i in data)
    ]
    
  ##### Validation #####
  
  def validate(self, validator=None):
    validator = validation.make_validator(validator)
    self.log('Loading...')
    self.preload()
    # required
    required = [
      'agency', 
      'stops', 
      'routes', 
      'trips',
      'stop_times',
      'calendar'
    ]
    for f in required:
      self.log("Validating required file: %s"%f)
      data = self.read(f)
      for i in data:
        i.validate(validator=validator)
        i.validate_feed(validator=validator)  
    # optional
    optional = [
      'calendar_dates',
      'fare_attributes',
      'fare_rules',
      'shapes',
      'frequencies',
      'transfers',
      'feed_info'
    ]
    for f in optional:
      self.log("Validating optional file: %s"%f)
      try:
        data = self.read(f)
      except KeyError, e:
        data = []
      for i in data:
        i.validate(validator=validator)
        i.validate_feed(validator=validator)  
    return validator

  def validate_feedvalidator(self, validator=None, report='report.html'):
    validator = validation.make_validator(validator)
    p = subprocess.Popen(
      [
        'feedvalidator.py',
        '--memory_db',
        '--noprompt',
        '--output',
        report,
        self.filename
      ],
      stdout=subprocess.PIPE, 
      stderr=subprocess.PIPE
    )
    stdout, stderr = p.communicate()
    returncode = p.returncode
    with validator(self):
      errors = [i for i in stdout.split('\n') if i.startswith('ERROR:')]
      if returncode:
        raise validation.ValidationError('Errors reported by feedvalidator.py; see %s for details'%report)



"""GTFS Stop entity."""
import entity
import validation

class Stop(entity.Entity):  
  """GTFS Stop entity."""
  ENTITY_TYPE = 's'
  KEY = 'stop_id'
  REQUIRED = [
    'stop_id',
    'stop_name',
    'stop_lat',
    'stop_lon'     
  ]
  OPTIONAL = [
    'stop_code',
    'stop_desc',
    'zone_id',
    'stop_url',
    'location_type',
    'parent_station',
    'stop_timezone',
    'wheelchair_boarding'
  ]
  
  def id(self):
    return self.get('stop_id')

  def name(self):
    return self.get('stop_name')
    
  def point(self):
    try:
      return float(self.get('stop_lon')), float(self.get('stop_lat'))
    except ValueError, e:
      return None

  def bbox(self):
    c = self.point()
    return [c[0], c[1], c[0], c[1]]

  def json(self):
    return {
      'type': 'Feature',
      'name': self.name(),
      'properties': self.data,
      'bbox': self.bbox(),
      'geometry': self.geometry()
    }
  
  def geometry(self):
    return {
      "type": 'Point',
      "coordinates": self.point(),
    }

  # Stop methods.
  def routes(self):
    serves = set()
    for stop_time in self._parents:
      for trip in stop_time._parents:
        serves |= trip._parents
    return serves

  def location_type(self):
    value = self.get('location_type')
    if value:
      value = int(value)
    else:
      value = 0
    return value
    
  ##### Validation #####
  def validate(self, validator=None):
    validator = super(Stop, self).validate(validator)
    # Required
    with validator(self): 
      assert self.get('stop_id'), "Required: stop_id"
    with validator(self): 
      assert self.get('stop_name'), "Required: stop_name"
    with validator(self): 
      assert self.get('stop_lat'), "Required: stop_lat"
    with validator(self): 
      assert self.get('stop_lon'), "Required: stop_lon"
    with validator(self): 
      assert validation.valid_point(self.point()), "Invalid stop_lon/stop_lat"
    # TODO: Warnings: 
    #  - stop too close to 0,0
    # Optional
    with validator(self):
      if self.get('stop_desc') == self.get('stop_name'):
        raise validation.ValidationWarning("stop_desc and stop_name are the same")        
    with validator(self):
      if self.get('zone_id') and int(self.get('location_type') or 0):
        raise validation.ValidationWarning("A station cannot have a zone_id")
    with validator(self): 
      if self.get('stop_url'):
        assert validation.valid_url(self.get('stop_url')), "Invalid stop_url"
    with validator(self): 
      if self.get('location_type'):
        assert validation.valid_bool(self.get('location_type'), empty=True), \
          "Invalid location_type"
    with validator(self):
      if int(self.get('location_type') or 0) and self.get('parent_station'):
        assert not self.get('parent_station'), \
          "A station cannot contain another station"          
    with validator(self):
      if self.get('stop_timezone'):
        assert validation.valid_tz(self.get('stop_timezone')), "Invalid timezone"
    with validator(self): 
      if self.get('wheelchair_boarding'):
        assert validation.valid_int(self.get('wheelchair_boarding'), vmin=0, vmax=2, empty=True), \
          "Invalid wheelchair_boarding"
    with validator(self):
      if self.get('stop_code'):
        pass
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(Stop, self).validate_feed(validator)  
    with validator(self):
      if self.get('parent_station'):
        parent_station = self._feed.stop(self.get('parent_station'))
        assert parent_station, "Unknown parent_station"
        assert int(parent_station.get('location_type') or 0) == 1, \
          "Invalid parent_station, parent must have location_type set to 1."
    return validator
    


"""GTFS Transfers."""
import datetime

import entity
import geom
import util
import widetime
import validation

class Transfer(entity.Entity):
  REQUIRED = [
    'from_stop_id',
    'to_stop_id',
    'transfer_type'
  ]
  OPTIONAL = [
    'min_transfer_time'
  ]
  
  def validate(self, validator=None):
    validator = super(Transfer, self).validate(validator)
    # Required
    with validator(self):
      assert self.get('from_stop_id'), "Required: from_stop_id"
    with validator(self):
      assert self.get('to_stop_id'), "Required: to_stop_id"
    with validator(self):
      # field required, blank allowed.
      assert validation.valid_int(self.get('transfer_type'), vmin=0, vmax=3, empty=True), \
        "Invalid transfer_type"
    with validator(self):
      if self.get('min_transfer_time'):
        assert validation.valid_int(self.get('min_transfer_time'), vmin=0), \
        "Invalid min_transfer_time"
    return validator
    
  def validate_feed(self, validator=None):
    validator = super(Transfer, self).validate_feed(validator)
    with validator(self):
      assert self._feed.stop(self.get('from_stop_id')), "Unknown from_stop_id"
    with validator(self):
      assert self._feed.stop(self.get('to_stop_id')), "Unknown to_stop_id"
    return validator